@@ -39,13 +39,16 @@
 #include "champlain-debug.h"
 
 #include "champlain.h"
-#include "champlain-cache.h"
+#include "champlain-file-cache.h"
 #include "champlain-defines.h"
 #include "champlain-enum-types.h"
 #include "champlain-map-source.h"
 #include "champlain-marshal.h"
 #include "champlain-private.h"
 #include "champlain-zoom-level.h"
+#include "champlain-network-tile-source.h"
+#include "champlain-map-source-chain.h"
+#include "champlain-error-tile-source.h"
 
 #include <glib.h>
 #include <string.h>
@@ -76,6 +79,8 @@ struct _ChamplainMapSourceFactoryPrivate
 static ChamplainMapSource * champlain_map_source_new_generic (
      ChamplainMapSourceDesc *desc, gpointer data);
 
+static ChamplainMapSource * champlain_map_source_new_memphis (
+    ChamplainMapSourceDesc *desc, gpointer user_data);
 
 static void
 champlain_map_source_factory_get_property (GObject *object,
@@ -86,7 +91,7 @@ champlain_map_source_factory_get_property (GObject *object,
   //ChamplainMapSourceFactory *map_source_factory = CHAMPLAIN_MAP_SOURCE_FACTORY(object);
   //ChamplainMapSourceFactoryPrivate *priv = map_source_factory->priv;
 
-  switch(prop_id)
+  switch (prop_id)
     {
       default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
@@ -102,7 +107,7 @@ champlain_map_source_factory_set_property (GObject *object,
   //ChamplainMapSourceFactory *map_source_factory = CHAMPLAIN_MAP_SOURCE_FACTORY(object);
   //ChamplainMapSourceFactoryPrivate *priv = map_source_factory->priv;
 
-  switch(prop_id)
+  switch (prop_id)
     {
       default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
@@ -160,7 +165,7 @@ ChamplainMapSourceDesc OSM_MAPNIK_DESC =
   {
     CHAMPLAIN_MAP_SOURCE_OSM_MAPNIK,
     "OpenStreetMap Mapnik",
-    "CC-BY-SA 2.0 OpenStreetMap contributors",
+    "Map data is CC-BY-SA 2.0 OpenStreetMap contributors",
     "http://creativecommons.org/licenses/by-sa/2.0/",
     0,
     18,
@@ -175,7 +180,7 @@ ChamplainMapSourceDesc OSM_OSMARENDER_DESC =
   {
     CHAMPLAIN_MAP_SOURCE_OSM_OSMARENDER,
     "OpenStreetMap Osmarender",
-    "CC-BY-SA 2.0 OpenStreetMap contributors",
+    "Map data is CC-BY-SA 2.0 OpenStreetMap contributors",
     "http://creativecommons.org/licenses/by-sa/2.0/",
     0,
     17,
@@ -190,7 +195,7 @@ ChamplainMapSourceDesc OSM_CYCLEMAP_DESC =
   {
     CHAMPLAIN_MAP_SOURCE_OSM_CYCLE_MAP,
     "OpenStreetMap Cycle Map",
-    "CC-BY-SA 2.0 OpenStreetMap contributors",
+    "Map data is CC-BY-SA 2.0 OpenStreetMap contributors",
     "http://creativecommons.org/licenses/by-sa/2.0/",
     0,
     18,
@@ -205,7 +210,7 @@ ChamplainMapSourceDesc OSM_TRANSPORTMAP_DESC =
   {
     CHAMPLAIN_MAP_SOURCE_OSM_TRANSPORT_MAP,
     "OpenStreetMap Transport Map",
-    "CC-BY-SA 2.0 OpenStreetMap contributors",
+    "Map data is CC-BY-SA 2.0 OpenStreetMap contributors",
     "http://creativecommons.org/licenses/by-sa/2.0/",
     0,
     18,
@@ -248,6 +253,36 @@ ChamplainMapSourceDesc MFF_RELIEF_DESC =
     NULL
   };
 
+static
+ChamplainMapSourceDesc MEMPHIS_LOCAL_DESC =
+  {
+    CHAMPLAIN_MAP_SOURCE_MEMPHIS_LOCAL,
+    "OpenStreetMap Memphis Local Map",
+    "(CC) BY 2.0 OpenStreetMap contributors",
+    "http://creativecommons.org/licenses/by/2.0/",
+    12,
+    18,
+    CHAMPLAIN_MAP_PROJECTION_MERCATOR,
+    champlain_map_source_new_memphis,
+    "",
+    NULL
+  };
+
+static
+ChamplainMapSourceDesc MEMPHIS_NETWORK_DESC =
+  {
+    CHAMPLAIN_MAP_SOURCE_MEMPHIS_NETWORK,
+    "OpenStreetMap Memphis Network Map",
+    "(CC) BY 2.0 OpenStreetMap contributors",
+    "http://creativecommons.org/licenses/by/2.0/",
+    12,
+    18,
+    CHAMPLAIN_MAP_PROJECTION_MERCATOR,
+    champlain_map_source_new_memphis,
+    "",
+    NULL
+  };
+
 static void
 champlain_map_source_factory_init (ChamplainMapSourceFactory *factory)
 {
@@ -271,6 +306,10 @@ champlain_map_source_factory_init (ChamplainMapSourceFactory *factory)
 #endif
   champlain_map_source_factory_register (factory, &MFF_RELIEF_DESC,
       MFF_RELIEF_DESC.constructor, MFF_RELIEF_DESC.data);
+  champlain_map_source_factory_register (factory, &MEMPHIS_LOCAL_DESC,
+      MEMPHIS_LOCAL_DESC.constructor, MEMPHIS_LOCAL_DESC.data);
+  champlain_map_source_factory_register (factory, &MEMPHIS_NETWORK_DESC,
+      MEMPHIS_NETWORK_DESC.constructor, MEMPHIS_NETWORK_DESC.data);
 }
 
 /**
@@ -329,9 +368,34 @@ champlain_map_source_factory_create (ChamplainMapSourceFactory *factory,
         return desc->constructor (desc, desc->data);
       item = g_slist_next (item);
     }
+
   return NULL;
 }
 
+ChamplainMapSource * champlain_map_source_factory_create_cached_source (ChamplainMapSourceFactory *factory,
+    const gchar *id)
+{
+  ChamplainMapSourceChain *source_chain;
+  ChamplainMapSource *tile_source;
+  ChamplainMapSource *error_source;
+  ChamplainMapSource *file_cache;
+  guint tile_size;
+
+  tile_source = champlain_map_source_factory_create (factory, id);
+
+  tile_size = champlain_map_source_get_tile_size (tile_source);
+  error_source = CHAMPLAIN_MAP_SOURCE(champlain_error_tile_source_new_full (tile_size));
+
+  file_cache = CHAMPLAIN_MAP_SOURCE(champlain_file_cache_new ());
+
+  source_chain = champlain_map_source_chain_new ();
+  champlain_map_source_chain_push (source_chain, error_source);
+  champlain_map_source_chain_push (source_chain, tile_source);
+  champlain_map_source_chain_push (source_chain, file_cache);
+
+  return CHAMPLAIN_MAP_SOURCE(source_chain);
+}
+
 /**
  * champlain_map_source_factory_register:
  * @factory: A #ChamplainMapSourceFactory
@@ -366,7 +430,7 @@ static ChamplainMapSource *
 champlain_map_source_new_generic (
      ChamplainMapSourceDesc *desc, gpointer user_data)
 {
-  return CHAMPLAIN_MAP_SOURCE (champlain_network_map_source_new_full (
+  return CHAMPLAIN_MAP_SOURCE (champlain_network_tile_source_new_full (
       desc->id,
       desc->name,
       desc->license,
@@ -377,3 +441,36 @@ champlain_map_source_new_generic (
       desc->projection,
       desc->uri_format));
 }
+
+static ChamplainMapSource *
+champlain_map_source_new_memphis (ChamplainMapSourceDesc *desc,
+    gpointer user_data)
+{
+  ChamplainMapDataSource *map_data_source;
+
+  if (g_strcmp0 (desc->id, CHAMPLAIN_MAP_SOURCE_MEMPHIS_LOCAL) == 0)
+    {
+      map_data_source = CHAMPLAIN_MAP_DATA_SOURCE (champlain_local_map_data_source_new ());
+
+      /* Abuse the uri_format field to store an initial data path (optional) */
+      if (desc->uri_format && g_strcmp0 (desc->uri_format, "") != 0)
+        champlain_local_map_data_source_load_map_data (
+            CHAMPLAIN_LOCAL_MAP_DATA_SOURCE (map_data_source),
+            desc->uri_format);
+    }
+  else if (g_strcmp0 (desc->id, CHAMPLAIN_MAP_SOURCE_MEMPHIS_NETWORK) == 0)
+      map_data_source = CHAMPLAIN_MAP_DATA_SOURCE (champlain_network_map_data_source_new ());
+  else
+    return NULL;
+
+  return CHAMPLAIN_MAP_SOURCE (champlain_memphis_tile_source_new_full (
+      desc->id,
+      desc->name,
+      desc->license,
+      desc->license_uri,
+      desc->min_zoom_level,
+      desc->max_zoom_level,
+      256,
+      desc->projection,
+      map_data_source));
+}
