@@ -56,6 +56,17 @@ toggle_layer (GtkToggleButton *widget,
     }
 }
 
+gboolean
+mouse_click_cb (ClutterActor *actor, ClutterEvent *event, gpointer data)
+{
+    gdouble lat, lon;
+
+    champlain_view_get_coords_from_event (CHAMPLAIN_VIEW (data), event, &lat, &lon);
+    g_print ("Mouse click at: %f  %f\n", lat, lon);
+
+    return TRUE;
+}
+
 static void
 map_source_changed (GtkWidget *widget,
                     ChamplainView *view)
@@ -75,13 +86,8 @@ map_source_changed (GtkWidget *widget,
   ChamplainMapSourceFactory *factory = champlain_map_source_factory_dup_default ();
   source = champlain_map_source_factory_create (factory, id);
 
-  if (source != NULL)
-    {
-      g_object_set (G_OBJECT (view), "map-source", source, NULL);
-      g_object_unref (factory);
-    }
-
-  g_object_unref (source);
+  g_object_set (G_OBJECT (view), "map-source", source, NULL);
+  g_object_unref (factory);
 }
 
 static void
@@ -210,6 +216,9 @@ main (int argc,
 
   widget = gtk_champlain_embed_new ();
   view = gtk_champlain_embed_get_view (GTK_CHAMPLAIN_EMBED (widget));
+  clutter_actor_set_reactive (CLUTTER_ACTOR (view), TRUE);
+  g_signal_connect (view, "button-release-event", G_CALLBACK (mouse_click_cb), view);
+
 
   g_object_set (G_OBJECT (view),
       "scroll-mode", CHAMPLAIN_SCROLL_MODE_KINETIC,
@@ -261,6 +270,8 @@ main (int argc,
   gtk_container_add (GTK_CONTAINER (bbox), button);
 
   button = gtk_spin_button_new_with_range(0, 20, 1);
+  gtk_spin_button_set_value (GTK_SPIN_BUTTON (button),
+      champlain_view_get_zoom_level (view));
   g_signal_connect (button, "changed", G_CALLBACK (zoom_changed), view);
   g_signal_connect (view, "notify::zoom-level", G_CALLBACK (map_zoom_changed),
       button);
